unit dnPxTbl;

(*****************************************************************************
Copyright  1997-99 Dieter Nth
version    : 2.1
date       : 10.08.1999
description:

    Enhanced TTable which loads Paradox masks defined (with DBD) on table level
    or from a corresponding INI-file.
    Validates entries against Paradox Min/Max checks.

    1.03 Added:
    property TableLevel: integer;(public) sets TableLevel for Paradox/DBase/FoxPro-tables
      (this is an example how to use dnDBUtil.ChangeTableParams)
      allowed values (from BDE.INT):
      const
        TBLLEVEL3  = 3;  { Paradox level 3 and dBASE level 3+ }
        TBLLEVEL4  = 4;  { Paradox level 4 and dBASE level 4 }
        TBLLEVEL5  = 5;  { Paradox level 5 and dBASE/Win }
        TBLLEVEL7  = 7;  { Paradox level 7 and dBASE level 7 }
        FOXLEVEL25 = 25; { Fox Table (Ver 2.5) }

    procedure AddIndexDescending(const Name, Fields: string; Options: TIndexOptions;
                                 const Descending : Array of WordBool);
    Like Borland's AddIndex, but Indices may be descending/ascending at
    field level (works only with Level 7 tables --> prop TableLevel).
    e.g. AddIndexDescending('NewIndex', 'Field1;Field2;Field3',
           [ixUnique, ixDescending, ixCaseInsensitive], [false, true, false]);
         creates Index with Field1 asc, Field2 desc, Field3 asc.
         Don't forget to include ixDescending in Options...
    Hint: You can't use prop IndexFieldNames to switch to this kind of index
      (exception SNoIndexForFields), use IndexName instead.
*)

{$I dnInc.INC}

interface

uses
  Windows, BDE,
  SysUtils, Classes, DB, DBTables, Dialogs, Controls,
  dnPxMask, dnGlobal, dnPxFlds;


type
  BoolArray = packed array[0..DBIMAXFLDSINKEY - 1] of wordBool;

  TPxTable = class(TTable)
  private
    { Private-Deklarationen }
    FMasksModified: boolean;
    function GetTableLevel : word;
    procedure SetTableLevel(newLevel : word);

    function GetMask(const Index : string) : TPxMask;
    procedure SetMask(const index : string; value : TPxMask);
  protected
    { Protected-Deklarationen }
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    procedure PrepareCursor; override;

    function IsParadoxTable: Boolean;
    {*** copied from DBTables.Pas}
    function IsDBaseTable: Boolean;
    function GetTableTypeName: PChar;

    {*** copied from DBTables.Pas & enhanced: DescFields}
    procedure DecodeIndexDesc(const IndexDesc: IDXDesc;
      var Source, Name, Fields: string; var Options: TIndexOptions;
      var DescFields : BoolArray);
    procedure EncodeIndexDesc(var IndexDesc: IDXDesc; const Name, Fields: string;
                              Options: TIndexOptions; const DescFields : Array of WordBool);
  public
    { Public-Deklarationen }
    function QualifiedTableName: String;
    procedure MaskModified;
    procedure AddIndexDescending(const Name, Fields: string; Options: TIndexOptions;
                                 const DescFields : Array of WordBool);
    function ReadMasks: boolean;
    function WriteMasks: boolean;

    property PxMasks[const index : string] : TPxMask read GetMask write SetMask;
  published
    property TableLevel : word read GetTableLevel write SetTableLevel stored false;
  end;

  procedure LoadParadoxValChecks(Table: TTable);

implementation

uses
  DBConsts, TypInfo,
  dnUtil, dnDBUtil, dnPxCnst;

{ TPxTable }
const
  ValidFieldTypes = [ftString, ftSmallint, ftInteger, ftWord,
    ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime];

procedure TPxTable.MaskModified;
begin
  FMasksModified := true;
end;

procedure TPxTable.SetFieldData(Field: TField; Buffer: Pointer);
var
  Blank: LongBool;

  procedure CheckMinMax(Status: DBIResult);
  var
    VCD : VChkDesc;
    sMin, sMax : string;

    function MinMaxStr(HasVal : word; aVal : Pointer): string;
    begin
      case HasVal of
        0: result := '';
        1: result := BDEData2Variant(aVal, Field);
      else
        if Field.DataType = ftDate then
          result := SToday
        else
          result := SNow;
      end;
    end;

    procedure Error(aErrStr : string; const Args : array of const);
    begin
      SysUtils.Beep;
      DatabaseErrorFmt(Format(SFieldValueError, [string(BDEData2Variant(Buffer, Field)),
                                          Field.DisplayName]) + aErrStr, Args);
    end;

  begin
    if Status = DBIERR_NONE then exit;
    GetValCheck(Field, VCD);
    with VCD do begin
      case Status of
        DBIERR_MINVALERR,
        DBIERR_MAXVALERR:
          begin
            sMin := MinMaxStr(word(bHasMinVal), @VCD.aMinVal);
            sMax := MinMaxStr(word(bHasMaxVal), @VCD.aMaxVal);
            if (sMin <> '') and (sMax <> '') then
              Error(SFieldMinMaxError,
                [sMin, sMax])
            else
            if sMin <> '' then
              Error(SFieldMinError, [sMin])
            else
              Error(SFieldMaxError, [sMax]);
          end;
        DBIERR_REQDERR:
          DatabaseErrorFmt(SFieldRequired, [Field.DisplayName]);
        DBIERR_LOOKUPTABLEERR: //not yet implemented
      end;
    end;
  end;
begin
  with Field do
    if (State in [dsEdit, dsInsert]) and
       (FieldNo > 0) and
       (not ReadOnly) and
       (FieldKind <> fkInternalCalc) then
    begin
      CheckMinMax(DbiVerifyField(Handle, FieldNo, Buffer, Blank));
    end;
  inherited;
end;

procedure TPxTable.PrepareCursor;
begin
  inherited;
  try
    ReadMasks;
  except
  end;
  LoadParadoxValChecks(self);
end;

function TPxTable.QualifiedTableName: String;
begin
  result := dnDBUtil.QualifiedTableName(self);
end;

function TPxTable.ReadMasks: boolean;
begin
  result := false;
  if isDBaseTable or isParadoxTable then
    result := LoadPxMasksFromIni(self,
                ChangeFileExt(QualifiedTableName, PxMaskFileExt), true);
  if result then FMasksModified := false;
end;

function TPxTable.WriteMasks: boolean;
begin
  result := false;
  if (active or (TableName <> '')) and (isDBaseTable or isParadoxTable) then
    result := StorePxMasksToIni(self,
                ChangeFileExt(QualifiedTableName, PxMaskFileExt), true);
  if result then FMasksModified := false;
end;

procedure TPxTable.SetMask(const index : string; value : TPxMask);
var
  pxm : TPxMask;
begin
  pxm := PxMasks[index];
  if pxm <> nil then
  begin
    pxm.assign(value);
    FMasksModified := true;
  end;
end;

function TPxTable.GetMask(const Index : string) : TPxMask;
var
  i : LongInt;
begin
  result := nil;
  if GetOrdProperty(FieldByName(Index), 'PxMask', tkClass, i) then
    result := TPxMask(i);
end;

function TPxTable.IsParadoxTable: Boolean;
begin
  Result := (TableType = ttParadox) or
    (CompareText(ExtractFileExt(TableName), '.DB') = 0);
end;

function TPxTable.IsDBaseTable: Boolean;
begin
  Result := (TableType = ttDBase) or
    (CompareText(ExtractFileExt(TableName), '.DBF') = 0);
end;

function TPxTable.GetTableTypeName: PChar;
const
  Names: array[TTableType] of PChar =
    (szPARADOX, szPARADOX, szDBASE, {$IFDEF Ver120} szFoxPro, {$ENDIF}szASCII);
var
  TT: TTableType;
  Extension: string;
begin
  Result := nil;
  if not Database.IsSQLBased then
  begin
    TT := TableType;
    if TT = ttDefault then
    begin
      Extension := ExtractFileExt(TableName);
      if CompareText(Extension, '.DBF') = 0 then TT := ttDBase;
      if CompareText(Extension, '.TXT') = 0 then TT := ttASCII;
    end;
    Result := Names[TT];
  end;
end;

procedure TPxTable.DecodeIndexDesc(const IndexDesc: IDXDesc;
      var Source, Name, Fields: string; var Options: TIndexOptions;
      var DescFields : BoolArray);
var
  IndexOptions: TIndexOptions;
  I: Integer;
  SSource, SName: PChar;
begin
  with IndexDesc do
  begin
    if szTagName[0] = #0 then
    begin
      SName := szName;
      Source := '';
    end
    else begin
      SSource := szName;
      SName := szTagName;
      NativeToAnsi(nil, SSource, Source);
    end;
    NativeToAnsi(Locale, SName, Name);
    Name := ExtractFileName(Name);
    Source := ExtractFileName(Source);
    IndexOptions := [];
    if bPrimary then Include(IndexOptions, ixPrimary);
    if bUnique then Include(IndexOptions, ixUnique);
    if bDescending then Include(IndexOptions, ixDescending);
    if bCaseInsensitive then Include(IndexOptions, ixCaseInsensitive);
    if bExpIdx then
    begin
      Include(IndexOptions, ixExpression);
      NativeToAnsi(Locale, szKeyExp, Fields);
    end else
    begin
      Fields := '';
      for I := 0 to iFldsInKey - 1 do
      begin
        if I <> 0 then Fields := Fields + ';';
        Fields := Fields + FieldDefs[aiKeyFld[I] - 1].Name;
      end;
      DescFields := BoolArray(abDescending);
    end;
    Options := IndexOptions;
  end;
end;

procedure TPxTable.EncodeIndexDesc(var IndexDesc: IDXDesc;
  const Name, Fields: string; Options: TIndexOptions; const DescFields : Array of WordBool);
var
  Pos: Integer;
begin
  FillChar(IndexDesc, SizeOf(IndexDesc), 0);
  with IndexDesc do
  begin
    if IsDBaseTable then
      AnsiToNative(Locale, Name, szTagName, SizeOf(szTagName) - 1)
    else
      AnsiToNative(Locale, Name, szName, SizeOf(szName) - 1);
    bPrimary := ixPrimary in Options;
    bUnique := ixUnique in Options;
    bDescending := ixDescending in Options;
    bMaintained := True;
    bCaseInsensitive := ixCaseInsensitive in Options;
    if ixExpression in Options then
    begin
      bExpIdx := True;
      AnsiToNative(Locale, Fields, szKeyExp, SizeOf(szKeyExp) - 1);
    end else
    begin
      Move(DescFields, abDescending, sizeOf(WordBool) * (High(DescFields) - Low(DescFields)));
      Pos := 1;
      while (Pos <= Length(Fields)) and (iFldsInKey < 16) do
      begin
        aiKeyFld[iFldsInKey] :=
          FieldDefs.Find(ExtractFieldName(Fields, Pos)).FieldNo;
        if bDescending then
          abDescending[iFldsInKey] := DescFields[iFldsInKey];
        Inc(iFldsInKey);
      end;
    end;
  end;
end;

procedure TPxTable.AddIndexDescending(const Name, Fields: string;
  Options: TIndexOptions; const DescFields : Array of WordBool);
var
  STableName: array[0..DBIMAXTBLNAMELEN - 1] of Char;
  IndexDesc: IDXDesc;
  OldLocale, CursorLocale: TLocale;
  LName: string;
begin
  CursorLocale := nil;
  FieldDefs.Update;
  EncodeIndexDesc(IndexDesc, Name, Fields, Options, DescFields);
  if Active then
  begin
    CheckBrowseMode;
    CursorPosChanged;
    Check(DbiAddIndex(DBHandle, Handle, nil, nil, IndexDesc, nil));
  end
  else begin
    OldLocale := nil;
    LName := GetLanguageDriverName;
    if (LName <> '') and (OsLdLoadBySymbName(PChar(LName), CursorLocale) = 0) then
    begin
      OldLocale := Locale;
      SetLocale(CursorLocale);
    end;
    try
      SetDBFlag(dbfTable, True);
      try
        Check(DbiAddIndex(DBHandle, nil, AnsiToNative(DBLocale, TableName,
          STableName, SizeOf(STableName) - 1), GetTableTypeName,
          IndexDesc, nil));
      finally
        SetDBFlag(dbfTable, False);
      end;
    finally
      if CursorLocale <> nil then
      begin
        OsLdUnloadObj(CursorLocale);
        SetLocale(OldLocale);
      end;
    end;
  end;
{  FIndexDefs.FUpdated := False; this direct change is replaced by}
  DataEvent(dePropertyChange, 0);
end;

function TPxTable.GetTableLevel : word;
var
  iLen : word;
begin
  if (not active) or
     (DbiGetProp(hDBIObj(handle), curTABLELEVEL, @result, 2, iLen) <> DBIERR_NONE) then
    result := 0;
end;

procedure TPxTable.SetTableLevel(newLevel : word);
begin
  CheckActive;
  if (GetTableLevel <> newLevel) and (
       (not (csDesigning in ComponentState) or
       (MessageDlg(SChangeTableLevel, mtConfirmation, [mbYes, mbNo], 0) = mrYes))
      )
  then
    ChangeTableParams(self, Format('%s%s%d', [szCFGDRVLEVEL, '=', newLevel]));
end;

procedure LoadParadoxValChecks(Table: TTable);
var
  ValCheckDesc: VCHKDesc;
  CursorProps: CurProps;
  i : integer;
  s : string;
  pxm : TPxMask;
begin
  with Table do begin
    DbiGetCursorProps(Handle, CursorProps);
    for I := 1 to CursorProps.iValChecks do
    begin
      DbiGetVChkDesc(Handle, I, @ValCheckDesc);
      with ValCheckDesc do begin
        NativeToAnsi(Locale, @szPict, s);
        if (s <> '') and GetOrdProperty( FindField(FieldDefs[iFldNum - 1].Name),
                         'PxMask', tkClass, LongInt(pxm)) then
        begin
          exclude(pxm.FState, pxMaskFromParadox);
          Pxm.Mask := s;
          include(pxm.FState, pxMaskFromParadox);
        end;
      end;
    end;
  end;
end;

end.
